import itertools

from .backends import StatevectorBackend
from .register import Register


class QuantumRegister(Register):
    instance_counter = itertools.count()
    prefix = 'q'

    def __init__(self, size, name=None, backend=StatevectorBackend):
        super().__init__(size, name)
        self.backend = backend(self.size, self.name)

    def apply_gate(self, gate, *params, adjoint=False):
        if len(params) != gate.arity:
            raise Exception(
                'Gate {} expects {} parameters, got {}'.format(
                    gate.name, gate.arity, len(params)
                )
            )

        self.backend.apply_gate(gate, *params, adjoint=adjoint)

    def measure_all(self):
        return self.backend.measure_all()

    def measure(self, target):
        return self.backend.measure(target)

    def yield_state(self):
        return self.backend.yield_state()

    def to_dirac(self):
        return self.backend.to_dirac()



import numpy as np
import numpy.linalg as la
import functools
import itertools

from qsy import gates
from qsy.util import format_complex

from .backend import Backend


class StatevectorBackend(Backend):
    '''
    StatevectorBackend uses a vector to describe the state of the system.
    To evolve the system, vector matrix multiplication is used. It can simulate
    all supported clifford and non-clifford gates.
    '''

    def __init__(self, size, name):
        self.name = name
        self.size = size
        self.state_size = 2**size
        self.state = np.zeros(self.state_size)
        self.state[0] = 1  # Initialize to zero state

    def apply_gate(self, gate, *params, adjoint=False):
        if gate.arity == 1:
            target = params[0]
            self._apply_single_qubit_gate(gate, target, adjoint)
        else:
            *controls, target = params
            self._apply_controlled_gate(gate, controls, target, adjoint)

    def measure_all(self):
        probabilities = np.abs(self.state)**2

        measured = np.random.choice(len(probabilities), p=probabilities)
        self.state[:] = 0
        self.state[measured] = 1

        binary_measurement = format(measured, '0{}b'.format(self.size))
        return [int(x) for x in binary_measurement]

    def measure(self, target):
        step_size = self.state_size // 2**(target+1)
        n_steps = (self.state.size // step_size) // 2

        # Create masks to extract amplitudes where target qubit is zero and one.
        zeros_mask = np.tile(
            np.concatenate((np.ones(step_size), np.zeros(step_size))), n_steps
        )
        ones_mask = np.tile(
            np.concatenate((np.zeros(step_size), np.ones(step_size))), n_steps
        )

        zero_amplitudes = self.state * zeros_mask
        one_amplitudes = self.state * ones_mask
        probabilities = np.abs(zero_amplitudes + one_amplitudes)**2

        measured = np.random.choice(len(probabilities), p=probabilities)
        measured_value = 1 if one_amplitudes[measured] != 0.0 else 0

        if measured_value == 1:
            for i in zero_amplitudes.nonzero():
                self.state[i] = 0
        else:
            for i in one_amplitudes.nonzero():
                self.state[i] = 0

        self._normalize()

        return measured_value

    def yield_state(self):
        for i, amplitude in np.ndenumerate(self.state):
            yield i[0], amplitude

    def to_dirac(self):
        return ' '.join('{}|{:0{n:d}b}>'.format(format_complex(a), i, n=self.size)
                        for a, i in zip(self.state, itertools.count())
                        if not np.isclose(a, 0.0))

    def _apply_single_qubit_gate(self, gate, target, adjoint):
        self._check_in_range(target)

        operation = gate.adjoint_matrix if adjoint else gate.matrix

        transformation = [gates.I.matrix] * self.size
        transformation[target] = operation

        transformation_matrix = functools.reduce(np.kron, transformation)

        self._transform(transformation_matrix)

    def _apply_controlled_gate(self, gate, controls, target, adjoint):
        self._check_in_range(target)

        for control in controls:
            self._check_in_range(control)

        operation = gate.adjoint_matrix if adjoint else gate.matrix
        control_matrix = np.array([[float('nan'), 0],
                                   [0, 1]])
        transformation = []

        for i in range(self.size):
            if i in controls:
                transformation.append(control_matrix)
            elif i == target:
                transformation.append(operation)
            else:
                transformation.append(gates.I.matrix)

        transformation_matrix = functools.reduce(np.kron, transformation)

        for (i, j), value in np.ndenumerate(transformation_matrix):
            if np.isnan(value):
                if i == j:
                    transformation_matrix[i, j] = 1
                else:
                    transformation_matrix[i, j] = 0

        self._transform(transformation_matrix)

    def _transform(self, operation):
        self.state = np.dot(self.state, operation)

    def _normalize(self):
        self.state = self.state / la.norm(self.state)



from enum import Enum
from .error import QsyASMError


class Operation:
    # Gates
    GATES_START = 0
    I = 1
    X = 2
    Y = 3
    Z = 4

    H = 5
    S = 6
    CX = 7
    CY = 8
    CZ = 9

    T = 10
    RX = 11
    RY = 12
    RZ = 13
    CRX = 14
    CRY = 15
    CRZ = 16
    CCX = 17
    GATES_END = 18

    # Registers
    QR = 19
    CR = 20

    MEASURE = 21

    ERROR = 22


operations = {
    'i': Operation.I,
    'x': Operation.X,
    'y': Operation.Y,
    'z': Operation.Z,

    'h': Operation.H,
    's': Operation.S,
    'cx': Operation.CX,
    'cz': Operation.CZ,

    't': Operation.T,
    'rx': Operation.RX,
    'ry': Operation.RY,
    'rz': Operation.RZ,
    'crx': Operation.CRX,
    'cry': Operation.CRY,
    'crz': Operation.CRZ,
    'ccx': Operation.CCX,

    'qreg': Operation.QR,
    'creg': Operation.CR,

    'meas': Operation.MEASURE,

    'error': Operation.ERROR
}


class Instruction:
    def __init__(self, op, args, lineno, lexpos):
        self.op = op
        self.op_name = op
        self.args = args
        self.lineno = lineno
        self.lexpos = lexpos

        self.type = self._get_op_type()

        # only applicable for gates
        self.adjoint = False

    def is_gate(self):
        return self.type > Operation.GATES_START and self.type < Operation.GATES_END

    def toggle_adjoint(self):
        self.adjoint = not self.adjoint

    def _get_op_type(self):
        self.op_name = self.op

        if isinstance(self.op, tuple):
            # parameterized instruction like qreg[n] or c(x)
            self.op_name, _ = self.op

        if self.op_name not in operations:
            return Operation.ERROR

        return operations[self.op_name]

    def __repr__(self):
        return '{}<{} {}>'.format(self.__class__.__name__, self.op, self.args)



import ply.yacc as yacc
import itertools
import numpy as np

from . import tokens
from .instruction import Instruction
from .error import ParseError


class QsyASMParser:
    tokens = tokens
    precedence = (
        ('left', 'PLUS', 'MIN'),
        ('left', 'MUL', 'DIV'),
        ('left', 'POW'),
        ('right', 'UMIN')
    )
    variables = {
        'pi': np.pi
    }

    def __init__(self, **kwargs):
        self.parser = yacc.yacc(module=self, **kwargs)

    def p_program(self, p):
        '''program : program instruction
                   | instruction'''
        if len(p) == 2 and p[1]:
            p[0] = [p[1]]
        elif len(p) == 3:
            p[0] = p[1]

            if not p[0]:
                p[0] = []

            if p[2]:
                p[0] += [p[2]]

    def p_instruction(self, p):
        '''instruction : normal_instruction
                       | adjoint_instruction'''
        p[0] = p[1]

    def p_normal_instruction(self, p):
        '''normal_instruction : term argument_list
                              | param_term argument_list'''
        p[0] = Instruction(p[1], p[2], p.lineno(0), p.lexpos(0))

    def p_adjoint_instruction(self, p):
        'adjoint_instruction : ADJ normal_instruction'
        instr = p[2]

        if not instr.is_gate():
            raise ParseError('Invalid operation "{}" for adjoint'.format(instr.op_name),
                             p.lexpos(2), p.lineno(2))

        instr.toggle_adjoint()
        p[0] = instr

    def p_instruction_newline(self, p):
        'instruction : NEWLINE'
        pass

    def p_term(self, p):
        '''term : IDENT
                | INTEGER
                | lookup'''
        p[0] = p[1]

    def p_lookup(self, p):
        'lookup : IDENT LBRACKET INTEGER RBRACKET'
        p[0] = (p[1], p[3])

    def p_param_term(self, p):
        'param_term : IDENT LPAREN expression RPAREN'
        p[0] = (p[1], p[3])

    def p_argument_list(self, p):
        'argument_list : term'
        p[0] = [p[1]]

    def p_argument_list_args(self, p):
        'argument_list : term COMMA argument_list'
        p[0] = [p[1]]
        p[0] += p[3]

    def p_expression_bin_op(self, p):
        '''expression : expression PLUS expression
                      | expression MIN expression
                      | expression DIV expression
                      | expression POW expression
                      | expression MUL expression'''
        if p[2] == '+':
            p[0] = p[1] + p[3]
        elif p[2] == '-':
            p[0] = p[1] - p[3]
        elif p[2] == '/':
            if p[3] == 0.0:
                raise ParseError('Division by zero', p.lexpos(2), p.lineno(2))
            p[0] = p[1] / p[3]
        elif p[2] == '**':
            p[0] = p[1] ** p[3]
        elif p[2] == '*':
            p[0] = p[1] * p[3]

    def p_expression_group(self, p):
        'expression : LPAREN expression RPAREN'
        p[0] = p[2]

    def p_expression_integer(self, p):
        'expression : INTEGER'
        p[0] = p[1]

    def p_expression_float(self, p):
        'expression : FLOAT'
        p[0] = p[1]

    def p_expression_unary_min(self, p):
        'expression : MIN expression %prec UMIN'
        p[0] = -p[2]

    def p_expression_ident(self, p):
        'expression : IDENT'
        if p[1] not in self.variables:
            raise ParseError(
                'Undefined variable "{}"'.format(p[1]),
                p.lexpos(1), p.lineno(1)
            )

        p[0] = self.variables[p[1]]

    def p_error(self, p):
        raise ParseError('Unexpected "{}"'.format(p.type), p.lexpos, p.lineno)

    def parse(self, s):
        return self.parser.parse(s, debug=False, tracking=True)




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'leftPLUSMINleftMULDIVleftPOWrightUMINADJ COMMA DIV FLOAT IDENT INTEGER LBRACKET LPAREN MIN MUL NEWLINE PLUS POW RBRACKET RPARENprogram : program instruction\n                   | instructioninstruction : normal_instruction\n                       | adjoint_instructionnormal_instruction : term argument_list\n                              | param_term argument_listadjoint_instruction : ADJ normal_instructioninstruction : NEWLINEterm : IDENT\n                | INTEGER\n                | lookuplookup : IDENT LBRACKET INTEGER RBRACKETparam_term : IDENT LPAREN expression RPARENargument_list : termargument_list : term COMMA argument_listexpression : expression PLUS expression\n                      | expression MIN expression\n                      | expression DIV expression\n                      | expression POW expression\n                      | expression MUL expressionexpression : LPAREN expression RPARENexpression : INTEGERexpression : FLOATexpression : MIN expression %prec UMINexpression : IDENT'
    
_lr_action_items = {'NEWLINE':([0,1,2,3,4,5,10,11,12,13,14,15,16,17,28,37,],[5,5,-2,-3,-4,-8,-10,-11,-1,-14,-5,-9,-6,-7,-15,-12,]),'ADJ':([0,1,2,3,4,5,10,11,12,13,14,15,16,17,28,37,],[8,8,-2,-3,-4,-8,-10,-11,-1,-14,-5,-9,-6,-7,-15,-12,]),'IDENT':([0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,20,22,24,28,30,31,32,33,34,35,37,],[9,9,-2,-3,-4,-8,15,15,9,-9,-10,-11,-1,-14,-5,-9,-6,-7,21,15,21,21,-15,-13,21,21,21,21,21,-12,]),'INTEGER':([0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,22,24,28,30,31,32,33,34,35,37,],[10,10,-2,-3,-4,-8,10,10,10,-9,-10,-11,-1,-14,-5,-9,-6,-7,25,27,10,25,25,-15,-13,25,25,25,25,25,-12,]),'$end':([1,2,3,4,5,10,11,12,13,14,15,16,17,28,37,],[0,-2,-3,-4,-8,-10,-11,-1,-14,-5,-9,-6,-7,-15,-12,]),'LPAREN':([9,18,22,24,31,32,33,34,35,],[18,22,22,22,22,22,22,22,22,]),'LBRACKET':([9,15,],[19,19,]),'COMMA':([10,11,13,15,37,],[-10,-11,20,-9,-12,]),'FLOAT':([18,22,24,31,32,33,34,35,],[26,26,26,26,26,26,26,26,]),'MIN':([18,21,22,23,24,25,26,29,31,32,33,34,35,36,38,39,40,41,42,43,],[24,-25,24,32,24,-22,-23,32,24,24,24,24,24,-24,-21,-16,-17,-18,-19,-20,]),'RPAREN':([21,23,25,26,29,36,38,39,40,41,42,43,],[-25,30,-22,-23,38,-24,-21,-16,-17,-18,-19,-20,]),'PLUS':([21,23,25,26,29,36,38,39,40,41,42,43,],[-25,31,-22,-23,31,-24,-21,-16,-17,-18,-19,-20,]),'DIV':([21,23,25,26,29,36,38,39,40,41,42,43,],[-25,33,-22,-23,33,-24,-21,33,33,-18,-19,-20,]),'POW':([21,23,25,26,29,36,38,39,40,41,42,43,],[-25,34,-22,-23,34,-24,-21,34,34,34,-19,34,]),'MUL':([21,23,25,26,29,36,38,39,40,41,42,43,],[-25,35,-22,-23,35,-24,-21,35,35,-18,-19,-20,]),'RBRACKET':([27,],[37,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'program':([0,],[1,]),'instruction':([0,1,],[2,12,]),'normal_instruction':([0,1,8,],[3,3,17,]),'adjoint_instruction':([0,1,],[4,4,]),'term':([0,1,6,7,8,20,],[6,6,13,13,6,13,]),'param_term':([0,1,8,],[7,7,7,]),'lookup':([0,1,6,7,8,20,],[11,11,11,11,11,11,]),'argument_list':([6,7,20,],[14,16,28,]),'expression':([18,22,24,31,32,33,34,35,],[23,29,36,39,40,41,42,43,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> program","S'",1,None,None,None),
  ('program -> program instruction','program',2,'p_program','parser.py',26),
  ('program -> instruction','program',1,'p_program','parser.py',27),
  ('instruction -> normal_instruction','instruction',1,'p_instruction','parser.py',40),
  ('instruction -> adjoint_instruction','instruction',1,'p_instruction','parser.py',41),
  ('normal_instruction -> term argument_list','normal_instruction',2,'p_normal_instruction','parser.py',45),
  ('normal_instruction -> param_term argument_list','normal_instruction',2,'p_normal_instruction','parser.py',46),
  ('adjoint_instruction -> ADJ normal_instruction','adjoint_instruction',2,'p_adjoint_instruction','parser.py',50),
  ('instruction -> NEWLINE','instruction',1,'p_instruction_newline','parser.py',54),
  ('term -> IDENT','term',1,'p_term','parser.py',58),
  ('term -> INTEGER','term',1,'p_term','parser.py',59),
  ('term -> lookup','term',1,'p_term','parser.py',60),
  ('lookup -> IDENT LBRACKET INTEGER RBRACKET','lookup',4,'p_lookup','parser.py',64),
  ('param_term -> IDENT LPAREN expression RPAREN','param_term',4,'p_param_term','parser.py',68),
  ('argument_list -> term','argument_list',1,'p_argument_list','parser.py',72),
  ('argument_list -> term COMMA argument_list','argument_list',3,'p_argument_list_args','parser.py',76),
  ('expression -> expression PLUS expression','expression',3,'p_expression_bin_op','parser.py',81),
  ('expression -> expression MIN expression','expression',3,'p_expression_bin_op','parser.py',82),
  ('expression -> expression DIV expression','expression',3,'p_expression_bin_op','parser.py',83),
  ('expression -> expression POW expression','expression',3,'p_expression_bin_op','parser.py',84),
  ('expression -> expression MUL expression','expression',3,'p_expression_bin_op','parser.py',85),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','parser.py',100),
  ('expression -> INTEGER','expression',1,'p_expression_integer','parser.py',104),
  ('expression -> FLOAT','expression',1,'p_expression_float','parser.py',108),
  ('expression -> MIN expression','expression',2,'p_expression_unary_min','parser.py',112),
  ('expression -> IDENT','expression',1,'p_expression_ident','parser.py',116),
]



import itertools
from .register import Register


class ClassicalRegister(Register):
    instance_counter = itertools.count()
    prefix = 'c'

    def __init__(self, size, name=None):
        super().__init__(size, name)
        self.state = [0] * size

    def set_state(self, state):
        self.state = state

    def __getitem__(self, index):
        return self.state[index]

    def __setitem__(self, index, value):
        self.state[index] = value



import abc


class Backend(abc.ABC):
    @abc.abstractmethod
    def apply_gate(self, gate, *params, adjoint=False):
        raise NotImplementedError()

    @abc.abstractmethod
    def measure_all(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def measure(self, target):
        raise NotImplementedError()

    @abc.abstractmethod
    def yield_state(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def to_dirac(self):
        raise NotImplementedError()

    def _check_in_range(self, target):
        if target < 0 or target >= self.size:
            raise Exception(
                'Can\'t access {}[{}]: register index out of range (register size {})'.format(
                    self.name, target, self.size
                )
            )



import numpy as np
from .gate import Gate, C
from .pauli import X, Y, Z

H_MATRIX = 1/np.sqrt(2) * np.array([[1,  1],
                                    [1, -1]])
S_MATRIX = np.array([[1, 0],
                     [0, 1j]])

H = Gate('H', H_MATRIX, H_MATRIX, 1)

S = Gate('S', S_MATRIX, S_MATRIX.conj().T, 1)

CX = C(X)
CY = C(Y)
CZ = C(Z)



import numpy as np
from .gate import Gate, C, CC
from .pauli import X

T_MATRIX = np.array([[1, 0],
                     [0, np.exp(1j * np.pi / 4)]])

T = Gate('T', T_MATRIX, T_MATRIX.conj().T, 1)


def Rx(angle):
    matrix = np.array([[np.cos(angle/2), -1j * np.sin(angle/2)],
                       [-1j * np.sin(angle/2), np.cos(angle/2)]])
    return Gate('Rx', matrix, matrix.conj().T, 1)


def Ry(angle):
    matrix = np.array([[np.cos(angle/2), -np.sin(angle/2)],
                       [np.sin(angle/2), np.cos(angle/2)]])
    return Gate('Ry', matrix, matrix.conj().T, 1)


def Rz(angle):
    matrix = np.array([[1, 0],
                       [0, np.exp(1j * angle)]])
    return Gate('Rz', matrix, matrix.conj().T, 1)


def CRx(angle):
    return C(Rx(angle))


def CRy(angle):
    return C(Ry(angle))


def CRz(angle):
    return C(Rz(angle))


CCX = CC(X)



from setuptools import setup, find_packages
import re


with open('./qsy/__init__.py') as f:
    version, = re.findall('__version__ = \'(.*)\'', f.read())

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='qsy',
    version=version,
    packages=find_packages(),
    author='Steven Oud',
    author_email='soud@pm.me',
    description='A quantum computer state vector/stabilizer circuit simulator and assembly language',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/soudy/qsy',
    license='MIT',
    install_requires=['numpy>=1.13', 'ply>=3'],
    download_url='https://github.com/soudy/qsy/archive/v{}.tar.gz'.format(
        version),
    keywords=['quantum', 'computing', 'simulator', 'stabilizer', 'circuit',
              'assembly', 'chp', 'statevector'],
    entry_points={
        'console_scripts': ['qsyasm=qsyasm.cli:main']
    },
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Assemblers'
    ]
)
